// sun 11:39 pm

#define portOfPin(P)\
  (((P)>=0&&(P)<8)?&PORTD:(((P)>7&&(P)<14)?&PORTB:&PORTC))
#define ddrOfPin(P)\
  (((P)>=0&&(P)<8)?&DDRD:(((P)>7&&(P)<14)?&DDRB:&DDRC))
#define pinOfPin(P)\
  (((P)>=0&&(P)<8)?&PIND:(((P)>7&&(P)<14)?&PINB:&PINC))
#define pinIndex(P)((uint8_t)(P>13?P-14:P&7))
#define pinMask(P)((uint8_t)(1<<pinIndex(P)))

#define pinAsInput(P) *(ddrOfPin(P))&=~pinMask(P)
#define pinAsInputPullUp(P) *(ddrOfPin(P))&=~pinMask(P);digitalHigh(P)
#define pinAsOutput(P) *(ddrOfPin(P))|=pinMask(P)
#define digitalLow(P) *(portOfPin(P))&=~pinMask(P)
#define digitalHigh(P) *(portOfPin(P))|=pinMask(P)
#define isHigh(P)((*(pinOfPin(P))& pinMask(P))>0)
#define isLow(P)((*(pinOfPin(P))& pinMask(P))==0)
#define digitalState(P)((uint8_t)isHigh(P))

#include <SoftwareSerial.h>
#include <ESP8266wifi.h>
SoftwareSerial ESPserial(2, 3); // RX | TX

const int TRQ = A5;        // 4    
const int DIR = 6;        // 5   
const int STEP = 5;        // 6   
const int nSLEEP = 8;        // 7 real pin is 14
const int nFAULT = 9;        // 8 real pin is 15   
const int PIR = A4;        // 9   
const int ULTRA = A0;        // 10   
const int window_contact_close = 12;        // 11  
const int window_contact_open = 11;        // 12  
const int TEMP = A1;        // A0
const int MOISTURE = A3;        // 13  
String input = "";

int PIR_Status = 0, ultra_sonic_trip = 0, window_status = 1, close_window_status = 5, FAULT_status = 0,open_window_status = 5; // window is open initially
int not_first_time = 0;
unsigned long counter;
int first_time = 1;


float d_avg_1,d_avg_2,vel,T=15,d_delta;
float min_dist_error = 0.15, max_dist_error = 0.35;    // min and max allowed inches moved 
                                                    // should move .25 inches

/************************************
 ******* Start of Ryan's Code *******
 ************************************/
int commandnew = 3;
  
void setup() {
  // Setup pins
  pinMode(TRQ, OUTPUT); 
  pinMode(DIR, OUTPUT);
  pinMode(STEP, OUTPUT);
  pinMode(nSLEEP, OUTPUT);
  pinMode(nFAULT, INPUT);
  pinMode(PIR, INPUT);
  pinMode(window_contact_close, INPUT_PULLUP);  // window contact is 1 by default
  pinMode(window_contact_open, INPUT_PULLUP);  // window contact is 1 by default
  pinMode(TEMP, INPUT);
  pinMode(MOISTURE, INPUT);
  digitalWrite(TRQ,LOW); // High Torque Mode
  digitalWrite(nSLEEP,LOW); // Put motor driver to sleep
  vel = 331.5 + (.6*T);     // Set velocity based on temperature to be used for Ultrasonic
  Serial.begin(9600);
  ESPserial.begin(9600); 
  Serial.println("Starting in 10...");
  delay(10000);
  Serial.println("");
  Serial.println("Starting now");
}

void loop() {
  // initialize
  int rain;
  int closing_status, opening_status;
  float temp;
  int command = 2; // 0 - close command, 1 - close command, 2 - don't do anything
  
  // window_loc = window_loc_setup(); // are either of the switches currently pressed?
  Update_window_status ();            // window is either 0-closed, 1-open, or 2-midway

  // get sensor data
  rain = read_rain(); // rain = 0 - dry, 1 - raining
  temp = read_temp(); // temp in F?
  if(first_time)
  {
    command = comm_with_hub(rain,temp);
    command = 2;
    first_time = 0;
  }

  // rain sequence
  if(rain) 
  {
    Update_window_status ();
    command = comm_with_hub(rain,temp);
    if(window_status == 1 || window_status == 2) // if the window isn't closed, but also no errors
    {
      closing_status = close_the_window(); // close the window
      window_status = update_win_loc(closing_status,0); // updates window_loc
      if(window_status == 3 || window_status == 4 || window_status == 5)
      {
        Serial.println("there was an error");
        Serial.print("error type = ");
        Serial.print(window_status);
        Serial.println("\nkey: 3 - ultrasonic, 4 - nfault, 5 - IR");
        command = comm_with_hub(rain,temp);
        while(1); // for now, change later to send error report to hub, request restart
      }
    }
    // command = 0, 1, or 2, depending on if window should close, open, or do nothing
  }else
  {
    Update_window_status ();
    command = comm_with_hub(rain,temp); // tell hub temp, rain, window status 
                                        // find out if it needs to open or close based on temp
    Serial.print("command received: ");
    Serial.println(command);
    if(command == 0) // close the window
    {
      command = 2;
      if(window_status != 0) // is it already closed?
      {
        closing_status = close_the_window(); // close the window
        window_status = update_win_loc(closing_status,0); 
        if(window_status == 3 || window_status == 4 || window_status == 5)
        {
          Serial.println("there was an error");
          Serial.print("error type = ");
          Serial.print(window_status );
          Serial.println("\nkey:3 - ultrasonic, 4 - nfault, 5 - IR");
          command = comm_with_hub(rain,temp);
          while(1); // for now, change later to send error report to hub, request restart
        }
      }
    }
    else if(command == 1) // open the window
    {
      command = 2;
      if(window_status != 1) // is it already open?
      {
        opening_status = open_the_window(); // close the window
        window_status = update_win_loc(opening_status,1); // update location regarless of opening or closing
        if(window_status == 3 || window_status == 4 || window_status == 5)
        {
          Serial.println("there was an error");
          Serial.print("error type = ");
          Serial.print(window_status);
          Serial.println("\nkey: 3 - ultrasonic, 4 - nfault, 5 - IR");
          command = comm_with_hub(rain,temp);
          while(1); // for now, change later to send error report to hub, request restart
        }
      }
    }
  }
  delay(5000);
}

String read_data() {
  char inChar = 0;
  String outChar;
  while (Serial.available()>0) {
    while(inChar != '\n') {
      inChar = (char)Serial.read();
      if(inChar > 0 && inChar < 150){
        input += inChar;
      }
    }
    outChar = input;
    input = "";
  }
  return outChar;
}

// This function checks if it is raining. If rain, output=1. No rain, output=0.
int read_rain() 
{
  int rain_status;
  if (digitalRead(MOISTURE) == LOW) {
    rain_status = 1; // it is raining
  }
  else {
    rain_status = 0; // it is not raining
  }
  return rain_status;
}

// This function reads in the temperature. It outputs the temperature in celsius as a float.
float read_temp() 
{
  int a1,a2,a3,a4,a5;
  float output,analog_voltage,analog_total;
  a1 = analogRead(TEMP);
  delayMicroseconds(100);
  a2 = analogRead(TEMP);
  delayMicroseconds(100);
  a3 = analogRead(TEMP);
  delayMicroseconds(100);
  a4 = analogRead(TEMP);
  delayMicroseconds(100);
  a5 = analogRead(TEMP);
  analog_total = (float)(a1+a2+a3+a4+a5);
  analog_voltage = analog_total/5;
  output = (float)(analog_voltage); // just to make damn sure
  output = output*5*100/1023;           // ^^^ 5V, 10mV in formula, 2^10 bits
  output = output*9/5+32; // convert C to F 
  if(output > 99)
  {
    output = 99;
  }else if(output < 40)
  {
    output = 40;
  }
  return output;
}

int comm_with_hub(int rain, float temp)
{
  int command = 2;
  Serial.print("Rain: ");
  Serial.println(rain);
  Serial.print("Temp: ");
  Serial.println(temp);
  Serial.print("Window Status: ");
  if(window_status == 0){
    Serial.println("closed");
  }else if(window_status == 1){
    Serial.println("open");
  }else if(window_status == 2){
    Serial.println("midway");
  }else{
    Serial.println("error");
  }
  // send data
  String temp_s = (String)(int)round(temp)+49;
  char d1 = temp_s[0]+49;
  char d2 = temp_s[1]+49;
  char d3 = rain+97;
  char d4 = window_status+97;
  String data = (String)d1;
  data += (String)d2;
  data += (String)d3;
  data += (String)d4;
  //data += (String)'$';
  
  Serial.print("Sent data to ESP: ");
  Serial.println(data);
  ESPserial.println(data); //maybe should be .print or write?
  
  Serial.println("Waiting for command to open or close window from another device");
  while(!ESPserial.available()){
  }
  if(ESPserial.available())   
  {  
    String info = read_data_wifi();
    command = info[0]-'0';
  }
  return command; // 0 - close window, 1 - open window, 2 - do nothing
}

String read_data_wifi() {
  char inChar = 0;
  String outChar;
  while (ESPserial.available()>0) {
    while(inChar != '\n') {
      inChar = (char)ESPserial.read();
      if(inChar > 0 && inChar < 150){
        input += inChar;
      }
    }
    outChar = input;
    input = "";
  }
  return outChar;
}

int update_win_loc(int stat, int type)
{
  if(type){ // opening
    if(stat == 0){ // opened
      return 1; // location is opened
    }else if(stat == 1){
      return 4; // ultrasonic tripped
    }else if(stat == 3){
      return 3; // n fault
    }else{
      return 2;
    }
  }
  else if(!type){ // closing
    if(stat == 0){ // closed
      return 0; // location is closed
    }else if(stat == 1){
      return 4; // ultrasonic tripped
    }else if(stat == 3){
      return 3; // n fault
    }else{
      return 2;
    }
  }
}

int close_the_window()
{
  int close_window_status;
  not_first_time = 0;
  digitalWrite(nSLEEP,HIGH); // Turn on motor driver
  do {   
    close_window_status = Close_Window ();
  }while(close_window_status == 2);
  
  digitalWrite(nSLEEP,LOW); // Put motor driver to sleep
  return close_window_status;
}

int open_the_window()
{
  int open_window_status;
  not_first_time = 0;
  digitalWrite(nSLEEP,HIGH); // Turn on motor driver
  do {  
    open_window_status = Open_Window ();
  }while(open_window_status == 2);
   
  digitalWrite(nSLEEP,LOW); // Put motor driver to sleep
  return open_window_status;
}

/**********************************
 ******* End of Ryan's Code *******
 **********************************/
 
// This function is used to operate the motor at 1kHz
void PWM_1k () {
  counter = micros(); 
  do {
    digitalHigh(STEP);
    } while (micros() - counter < 100); // Approximately 10% duty cycle @ 1KHz
  do {
    digitalLow(STEP);
    } while (micros() - counter < 1000); // Approximately 10% duty cycle @ 1KHz
}


// This function is used for normal motor operation to ramp up the speed - takes 1.998 seconds
void PWM_ramp_up () {
  //for (long ramp_count = 1000000; ramp_count >= 1953; ramp_count = ramp_count/2) {
  for (long ramp_count = 100; ramp_count <= 1000; ramp_count = ramp_count + 100) {
  counter = micros(); 
  do {
    digitalHigh(STEP);
    } while (micros() - counter < 100000/ramp_count); // Approximately 10% duty cycle @ 1KHz
  do {
    digitalLow(STEP);
    } while (micros() - counter < 1000000/ramp_count); // Approximately 10% duty cycle @ 1KHz
  }
}

// This function is used to operate the motor at 1kHz while simultaneously updating the status of the PIR sensor,
// motor driver (nFAULT), and position of the window.
void PWM_1k_Safety () { 
  counter = micros(); 
  do {
    digitalHigh(STEP);
    } while (micros() - counter < 100); // Approximately 10% duty cycle @ 1KHz
  do {
    digitalLow(STEP);
    Update_PIR_Status ();
    Update_nFAULT_Status ();
    Update_window_status ();
    } while (micros() - counter < 300); // Approximately 10% duty cycle @ 1KHz
  
  do {
  digitalLow(STEP);
  } while (micros() - counter < 1000); // Approximately 10% duty cycle @ 1KHz
}


// This function is used to take a distance measurement from the ultrasonic sensor
float Check_ULTRA () { 

  unsigned long pulse_return=0, pulse_end=0; 
  float total_time=0, out=0;
  
  pinMode(ULTRA, OUTPUT);
 
  // Output 5us pulse
  digitalWrite(ULTRA, HIGH);
  digitalWrite(ULTRA, LOW);
  
  // Setup pin as input
  pinMode(ULTRA, INPUT);
  
  // Measure width of recieved pulse
  while (digitalState(ULTRA) == LOW) {
    pulse_return = micros();
  } // Waiting until the pin goes HIGH
  while (digitalState(ULTRA) == HIGH) {
    pulse_end = micros();
  } // Counting time until the pin goes LOW
  total_time = (float) (pulse_end - pulse_return);
  
  // Final calculated distance
  out = (vel*total_time/50800);
  return out;

}



// This function is used to take 5 distance measurements via the ultrasonic sensor and output an average distance value
float ULTRA_Distance () {
  float d1=0,d2=0,d3=0,d4=0,d5=0,d_out=0;
  d1 = Check_ULTRA(); // Check once per second
  Serial.print("\nd1 = ");
  Serial.print(d1);
  delay(100);
  d2 = Check_ULTRA(); // Check once per second
  Serial.print("\nd2 = ");
  Serial.print(d2);
  delay(100);
  d3 = Check_ULTRA(); // Check once per second
  Serial.print("\nd3 = ");
  Serial.print(d3);
  delay(100);
  d4 = Check_ULTRA(); // Check once per second
  Serial.print("\nd4 = ");
  Serial.print(d4);
  delay(100);
  d5 = Check_ULTRA(); // Check once per second
  Serial.print("\nd5 = ");
  Serial.print(d5);
  Serial.print("\n");
  d_out = ((d1+d2+d3+d4+d5)/5);
  return d_out;
}

// This function is used to check the status of the PIR sensor (HIGH/LOW)
void Update_PIR_Status () {
  if (digitalState(PIR) == HIGH) {
      PIR_Status = 1;   // PIR sensor is tripped
    }
    else {
      PIR_Status = 0;   // No motion detected
    }
}


// This function is used to temporarily stop the motor when the PIR sensor is tripped
void Check_PIR () {
  while ( PIR_Status == 1) {  // Stop motor until PIR sensor goes low
    Update_PIR_Status ();
  }
}

// This function is used to determine the position of the window (fully open, fully closed, or partially open)
void Update_window_status () {
  if ( (digitalState(window_contact_open) == LOW) && (digitalState(window_contact_close) == HIGH) ) {
      window_status = 1; // window fully open
    }
  else if ( (digitalState(window_contact_open) == HIGH) && (digitalState(window_contact_close) == LOW)){
      window_status = 0; // window closed
    }
  else {
      window_status = 2; // window is somewhere in between
  }
}



// This function is used to check the FAULT status of the motor driver
void Update_nFAULT_Status () {
  if (digitalState(nFAULT) == HIGH) {
      FAULT_status = 0;   // No Fault
    }
    else {
      FAULT_status = 1;   // Fault
    }
}

// This function is used for normal motor operation. It ramps up the speed first. After, it runs normally for 49ms. Then, it checks the PIR sensor while still running for 1ms. 
// It repeats 268 times to approximately move .25 inches
void Run_Motor_Close () {
  PWM_ramp_up (); // takes 29.29 ms to ramp up
  for(int i=0; i<268; i++) { // ~takes 13.4 seconds
    for(int j = 0; j<49; j++) { 
      PWM_1k(); // 1kHz for 49ms
    }
    PWM_1k_Safety();  // Check PIR,nFAULT,window_status at 1kHz for 1ms
    Check_PIR ();
    if (window_status == 0 || FAULT_status == 1) { // Exit loop if the window is already closed
      break;
    }
  }
}

// This function is used for normal motor operation. It ramps up the speed first. After, it runs normally for 49ms. Then, it checks the PIR sensor while still running for 1ms. 
// It repeats 268 times to approximately move .25 inches
void Run_Motor_Open () {
  PWM_ramp_up (); // takes 29.29 ms to ramp up
  for(int i=0; i<268; i++) {  // ~takes 13.4 seconds
    for(int j = 0; j<49; j++) { 
      PWM_1k(); // 1kHz for 49ms
    }
    PWM_1k_Safety(); // Check PIR,nFAULT,window_status at 1kHz for 1ms
    Check_PIR ();
    if (window_status == 1 || FAULT_status == 1) { // Exit loop if the window is already open
      break;
    }
  }
}



// This function is used to close the window
int Close_Window () {
  digitalWrite(DIR,LOW); // CW Rotation
  Run_Motor_Close ();
  if (window_status == 0) {
        return 0;
      }
  if (FAULT_status == 1) {
        return 3;
      }
  d_avg_1 = ULTRA_Distance ();
  Serial.print("\n Distance 1 = \n");
  Serial.print(d_avg_1);
  if (window_status == 0) {
        return 0;
      }
  if (FAULT_status == 1) {
        return 3;
      }

  if (not_first_time) { // don't do this the first time
    
    d_delta = d_avg_1 - d_avg_2;
    Serial.print("\n Delta = \n");
    Serial.print(d_delta);
    if ( ((d_delta > max_dist_error) || (d_delta < min_dist_error)) && ultra_sonic_trip == 0) {  // If delta_d is too large/small for the first time, wait 5 seconds and try again
        ultra_sonic_trip = 1;
        Serial.print("\n Ultrasonic Tripped");
        delay(5000); // Wait 5 seconds
    }
    else if ( ((d_delta > max_dist_error) || (d_delta < min_dist_error)) && ultra_sonic_trip == 1) { // If delta_d is too large/small for the second time, stop motor indefinitely
      Serial.print("\n Ultrasonic Tripped Again - Window Stopped");
      return 1; // STOP WINDOW
    }
    else {
      ultra_sonic_trip = 0;
    }
  }
  else {
    not_first_time = 1;      
  }


  Run_Motor_Close ();
  if (window_status == 0) {
        return 0;
      }
  if (FAULT_status == 1) {
        return 3;
      }
  d_avg_2 = ULTRA_Distance ();
  Serial.print("\n Distance 2 = ");
  Serial.print(d_avg_2);
  if (window_status == 0) {
        return 0;
      }
  if (FAULT_status == 1) {
        return 3;
      }
  d_delta = d_avg_2 - d_avg_1;
  

  Serial.print("\n Delta = ");
  Serial.print(d_delta);
  Serial.print("\n\n\n");
  
  
  if ( ((d_delta > max_dist_error) || (d_delta < min_dist_error)) && ultra_sonic_trip == 0) {  // If delta_d is too large/small for the first time, wait 5 seconds and try again
    ultra_sonic_trip = 1;
    Serial.print("\n Ultrasonic Tripped");
    delay(5000); // Wait 5 seconds
  }
  else if ( ((d_delta > max_dist_error) || (d_delta < min_dist_error)) && ultra_sonic_trip == 1) { // If delta_d is too large/small for the second time, stop motor indefinitely
    Serial.print("\n Ultrasonic Tripped Again - Window Stopped");
    return 1; // STOP WINDOW
  }
  else {
    ultra_sonic_trip = 0;
  }
  return 2;
}


// This function is used to open the window
int Open_Window () {
  digitalWrite(DIR,HIGH); // CCW Rotation
  Run_Motor_Open ();
  if (window_status == 1) {
        return 0;
      }
  if (FAULT_status == 1) {
        return 3;
      }
  d_avg_1 = ULTRA_Distance ();
  if (window_status == 1) {
        return 0;
      }
  if (FAULT_status == 1) {
        return 3;
      }

  
if (not_first_time) {
  
  d_delta = d_avg_2 - d_avg_1;
  Serial.print("\n Delta = \n");
  Serial.print(d_delta);
  if ( ((d_delta > max_dist_error) || (d_delta < min_dist_error)) && ultra_sonic_trip == 0) {  // If delta_d is too large/small for the first time, wait 5 seconds and try again
      ultra_sonic_trip = 1;
      Serial.print("\n Ultrasonic Tripped");
      delay(5000); // Wait 5 seconds
    }
    else if ( ((d_delta > max_dist_error) || (d_delta < min_dist_error)) && ultra_sonic_trip == 1) { // If delta_d is too large/small for the second time, stop motor indefinitely
      Serial.print("\n Ultrasonic Tripped Again - Window Stopped");
      return 1; // STOP WINDOW
    }
    else {
      ultra_sonic_trip = 0;
    }
}
else {
  not_first_time = 1;      
}



      
  Run_Motor_Open ();
  if (window_status == 1) {
        return 0;
      }
  if (FAULT_status == 1) {
        return 3;
      }
  d_avg_2 = ULTRA_Distance ();
  if (window_status == 1) {
        return 0;
      }
  if (FAULT_status == 1) {
        return 3;
      }
  d_delta = d_avg_1 - d_avg_2;
  if ( ((d_delta > max_dist_error) || (d_delta < min_dist_error)) && ultra_sonic_trip == 0) {  // If delta_d is too large/small for the first time, wait 5 seconds and try again
    ultra_sonic_trip = 1;
    Serial.print("\n Ultrasonic Tripped");
    delay(5000); // Wait 5 seconds
  }
  else if ( ((d_delta > max_dist_error) || (d_delta < min_dist_error)) && ultra_sonic_trip == 1) { // If delta_d is too large/small for the second time, stop motor indefinitely
    Serial.print("\n Ultrasonic Tripped Again - Window Stopped");
    return 1; // STOP WINDOW
  }
  else {
    ultra_sonic_trip = 0;
  }
  return 2;
}
