#include <SoftwareSerial.h>

SoftwareSerial espHub(2,3);

// I/O Pins
const int temp_pin = A2;  //temperature output
const int wifi_rst = A5;  //WiFi Reset

//Variables
int auto_control;       //0 for auto control off, 1 for auto control on
int outdoor_temp = 70;
float indoor_temp;
int min_temp = 60;
int max_temp = 80;  
int window_pos;  //0 for closed, 1 for open, 2 for in between
int rain_status;        //0 for no rain, 1 for rain
int first_time = 1;     //Stays high until first update
char buf;                //buffer for received data
int comm_status = 0;    //indicates device for communication
String window_input = "";
String app_input = "";
bool newData = false;
int action = 0;       //0 means nothing, 1 means cool, 2 means warm
int manual_move = 0;   //0 means nothing, 1 means close, 2 means open
int counter = 0;
int app_info_ready = 1;  //goes high when received data from app
int window_info_ready = 0;  //goes high when received data from window
int update_from_window = 0; //goes high when window string is in
int update_from_app = 0;
int threshold_diff = 2;

void setup() {
  pinMode(temp_pin, INPUT); 
  pinMode(wifi_rst, OUTPUT);
  auto_control = 1;
  Serial.begin(9600);
  //while(!Serial) {
  //}
  espHub.begin(9600);
  //delay(10000);
  //window_update();
  //app_update();
  indoor_temp = read_temp();
}

void loop() {

  //update indoor temperature
  //if(counter>5000) {
  indoor_temp = read_temp();
    //counter = 0;
  //}
  //counter = counter+1;

  //control algorithm
  if(app_info_ready && window_info_ready) {
    // start response edit here
    String response = "";
    String info_to_app = " Temp indoors: ";
    info_to_app += indoor_temp;
    info_to_app += " Temp outdoors: ";
    info_to_app += outdoor_temp;
    info_to_app += " Rain status: ";
    info_to_app += rain_status;
    info_to_app += " Window status (0 - closed, 1 - open): ";
    info_to_app += window_pos;
    info_to_app += " Min temp in comfort range: ";
    info_to_app += min_temp;
    info_to_app += " Max temp in comfort range: ";
    info_to_app += max_temp;
    info_to_app += " Automatic control status: ";
    info_to_app += auto_control;
    // end response edit
    if(auto_control) {
      if(!rain_status) {
        if(indoor_temp<min_temp) {
          action = 2;   // warm up
        }
        else if(indoor_temp>max_temp) {
          action = 1;   //cool down
        }
        else {
          action = 0;
        }
        //Currently Closed
        if(window_pos == 0) {
          if(outdoor_temp>(indoor_temp+threshold_diff) && action == 2) {
            //Serial.print("\nOpen Window");
            response = "1";
            response += info_to_app;
            espHub.println(response);  //Tell Window to open
            //Serial.print('\n');
            //window_pos = 1;
          }
          else if(outdoor_temp<(indoor_temp-threshold_diff) && action == 1) {
            //Serial.print("\nOpen Window");
            response = "1";
            response += info_to_app;
            espHub.println(response);  //Tell Window to open
            //Serial.print('\n');
            //window_pos = 1;
          }
          else {
            response = "2";
            response += info_to_app;
            espHub.println(response); //Do nothing
          }
        }
        //Currently Open
        if(window_pos == 1) {
          if(outdoor_temp<(indoor_temp-threshold_diff) && action == 2) {
            //Serial.print("\nClose Window");
            //window_pos = 0;
            response = "0";
            response += info_to_app;
            espHub.println(response);  //Tell Window to close
            //Serial.print('\n');
          }
          else if(outdoor_temp>(indoor_temp+threshold_diff) && action == 1) {
            //Serial.print("\nClose Window");
            //window_pos = 0;
            response = "0";
            response += info_to_app;
            espHub.println(response);  //Tell Window to close
            //Serial.print('\n');
          }
          else {
            response = "2";
            response += info_to_app;
            espHub.println(response); //Do nothing
          }
        }
        //if window is midway
        else {
            response = "1";
            response += info_to_app;
            espHub.println(response);  //Tell Window to open
        }
        
      }
      else {
        //If currently open and raining
        if(window_pos == 1 || window_pos == 2) {
          //window_pos = 0;
          //Serial.print("\nClose Window");
          //Serial.print("\nSleep");
            response = "0";
            response += info_to_app;
            espHub.println(response);  //Tell Window to close
          //Serial.print('\n');
        }
        else {
            response = "2";
            response += info_to_app;
            espHub.println(response); //Do nothing
          }
      }
    }
    //if auto_control off
    else {
      if(manual_move) {
            response = "1";
            response += info_to_app;
            espHub.println(response);  //open window manually
      }
      else if(!manual_move) {
            response = "0";
            response += info_to_app;
            espHub.println(response);  //close window manually
      }
    }
  }
  
  //Serial.print("\n1 for window, 2 for app status");
  //while(!Serial.available()){
  //}
  read_data();
  //int Update = read_data(); //read off
  if(update_from_window) {
    window_update();
    window_status();
    update_from_window = 0;
  }
  if(update_from_app) {
    app_update();
    app_status();
    update_from_app = 0;
  }


  

}

void read_data() {
  int window_done = 0;
  int app_done = 0;
  char inChar = 0;
  char garbage;
  app_input = "";
  window_input = "";
  if (espHub.available()>0) {
    //Serial.print("Reading Data\n");
    while(1) {

      // get the new byte:
      inChar = (char)espHub.read();
      // add it to the inputString:
      if(isDigit(inChar)){
        app_input += inChar;
      }
      else if(isAlpha(inChar)) {
        window_input += inChar;
      }
      else if(inChar == '$') {
        window_done = 1;
       // garbage = (char)Serial.read();
      }
      else if(inChar == '@') {
        app_done = 1;
       // garbage = (char)Serial.read();
      }
      
      if(window_done && app_input == "") {
        update_from_window = 1;
        //garbage = (char)espHub.read();
        Serial.print("\n");
        Serial.print(window_input);
        window_input = string_to_int(window_input);
        return;
      }
      else if(app_done && window_input == "") {
        update_from_app = 1;
        //garbage = (char)espHub.read();
        Serial.print("\n");
        Serial.print(app_input);
        return;
      }
      else if(window_done && app_done) {
        update_from_window = 1;
        update_from_app = 1;
        Serial.print("\nboth");
        //garbage = (char)Serial.read();
        return;
      }

    }

  }

}

void window_update (){
  Serial.print("\nIn Window Update");
  //Serial.print("\nWindow Position (Closed, Open, Midway): ");
  //while(!Serial.available()){
  //}
  //read_data();

  outdoor_temp = char_to_int(window_input[0],window_input[1]);
  window_info_ready = 1;
  
  int temp = int(window_input[2])-'0';

  if(temp == 0 || temp == 1) {
    rain_status = temp;
  }
  else{
    error_msg();
    return;
  }

  temp = int(window_input[3])-'0';
  
  if(temp == 0 || temp == 1 || temp == 2) {
    window_pos = temp;
  }
  else{
    error_msg();
    return;
  }

  
  
}

void app_update() {
  Serial.print("\nIn App Update");
  //while(!Serial.available()){
  //}
  //read_data();
  min_temp = char_to_int(app_input[0],app_input[1]);
  max_temp = char_to_int(app_input[2],app_input[3]);

  int temp = int(app_input[4])-'0';
  if(temp == 0 || temp == 1) {
    auto_control = temp;
  }
  else{
    error_msg();
    return;
  }

  temp = int(app_input[5])-'0';
  if(temp == 0 || temp == 1 || temp == 2) {
    manual_move = temp;
  }
  else{
    error_msg();
    return;
  }

  app_info_ready = 1;
  
}

void window_status() {

  Serial.print("\nwindow_pos = ");
  Serial.print(window_pos);
  Serial.print("\nrain_status = ");
  Serial.print(rain_status);
  Serial.print("\noutdoor_temp = ");
  Serial.print(outdoor_temp);
  //int num = 2;
  //espHub.println(2);
  
}

void app_status() {

  Serial.print("\nmin_temp = ");
  Serial.print(min_temp);
  Serial.print("\nmax_temp = ");
  Serial.print(max_temp);
  Serial.print("\nauto_control = ");
  Serial.print(auto_control);
  Serial.print("\nmanual_move = ");
  Serial.print(manual_move);
  Serial.print("\nindoor_temp = ");
  Serial.print(indoor_temp);
  
}

void error_msg() {
  Serial.print("\nInvalid Input");
}

float read_temp() 
{
  int a1,a2,a3,a4,a5;
  float output,analog_voltage,analog_total;
  a1 = analogRead(temp_pin);
  delayMicroseconds(100);
  a2 = analogRead(temp_pin);
  delayMicroseconds(100);
  a3 = analogRead(temp_pin);
  delayMicroseconds(100);
  a4 = analogRead(temp_pin);
  delayMicroseconds(100);
  a5 = analogRead(temp_pin);
  analog_total = (float)(a1+a2+a3+a4+a5);
  analog_voltage = analog_total/5;
  output = (float)(analog_voltage);
  output = output*5*100/1023;           // ^^^ 5V, 10mV in formula, 2^10 bits
  output = output*9/5+32; // convert C to F
  return output;
}

int char_to_int(char A,char B){
  
  String temp_str = "";
  temp_str += A;
  temp_str += B;
  return temp_str.toInt();
  
}

//converts string of lowercase letters into numbers
//outputs as an integer
String string_to_int(String in) {
  //Serial.print(in);
  //Serial.print('\n');
  int i;
  for(i=0;i<in.length();i++) {
    in[i] = in[i]-49;
  }
  //Serial.print(in);
  //Serial.print('\n');
  return in;
  
}

